@@ -54,6 +54,8 @@
 #include <linux/connector.h>
 #include <linux/cn_proc.h>
 #include <linux/un.h>
+#include <pwd.h>
+#include <grp.h>
 
 #define NUM_PER_REALLOCATIOM	(100)
 
@@ -66,6 +68,12 @@ int logfacility;
 /* Current log level */
 int loglevel;
 
+/* Owner of the socket, -1 means no change */
+uid_t socket_user = -1;
+
+/* Owner of the socket, -1 means no change */
+gid_t socket_group = -1;
+
 /**
  * Prints the usage information for this program and, optionally, an error
  * message.  This function uses vfprintf.
@@ -94,6 +102,10 @@ static void usage(FILE* fd, const char* msg, ...)
 		"    -n           | --nodaemom          don't fork daemon\n"
 		"    -d           | --debug             same as -v -v -n -f -\n"
 		"    -Q           | --nolog             disable logging\n"
+		"    -u <user>    | --socket-user=<user> set "
+			CGRULE_CGRED_SOCKET_PATH " socket user\n"
+		"    -g <group>   | --socket-group=<group> set "
+			CGRULE_CGRED_SOCKET_PATH " socket group\n"
 		"    -h           | --help              show this help\n\n"
 		);
 	va_end(ap);
@@ -161,13 +173,15 @@ static int cgre_store_parent_info(pid_t pid)
 	uptime_ns = ((__u64)tp.tv_sec * 1000 * 1000 * 1000 ) + tp.tv_nsec;
 
 	if (array_pi.index >= array_pi.num_allocation) {
-		array_pi.num_allocation += NUM_PER_REALLOCATIOM;
-		array_pi.parent_info = realloc(array_pi.parent_info,
-					sizeof(info) * array_pi.num_allocation);
-		if (!array_pi.parent_info) {
+		int alloc = array_pi.num_allocation + NUM_PER_REALLOCATIOM;
+		void *new_array = realloc(array_pi.parent_info,
+					  sizeof(info) * alloc);
+		if (!new_array) {
 			flog(LOG_WARNING, "Failed to allocate memory");
 			return 1;
 		}
+		array_pi.parent_info = new_array;
+		array_pi.num_allocation = alloc;
 	}
 	info = calloc(1, sizeof(struct parent_info));
 	if (!info) {
@@ -246,13 +260,15 @@ static int cgre_store_unchanged_process(pid_t pid, int flags)
 		return 0;
 	}
 	if (array_unch.index >= array_unch.num_allocation) {
-		array_unch.num_allocation += NUM_PER_REALLOCATIOM;
-		array_unch.proc = realloc(array_unch.proc,
-			sizeof(unchanged_pid_t) * array_unch.num_allocation);
-		if (!array_unch.proc) {
+		int alloc = array_unch.num_allocation + NUM_PER_REALLOCATIOM;
+		void *new_array = realloc(array_unch.proc,
+					  sizeof(unchanged_pid_t) * alloc);
+		if (!new_array) {
 			flog(LOG_WARNING, "Failed to allocate memory");
 			return 1;
 		}
+		array_unch.proc = new_array;
+		array_unch.num_allocation = alloc;
 	}
 	array_unch.proc[array_unch.index].pid = pid;
 	array_unch.proc[array_unch.index].flags = flags;
@@ -327,6 +343,12 @@ int cgre_process_event(const struct proc_event *ev, const int type)
 	switch (type) {
 	case PROC_EVENT_UID:
 	case PROC_EVENT_GID:
+		/*
+		 * If the unchanged process, the daemon should not change the
+		 * cgroup of the process.
+		 */
+		if (cgre_is_unchanged_process(ev->event_data.id.process_pid))
+			return 0;
 		pid = ev->event_data.id.process_pid;
 		break;
 	case PROC_EVENT_FORK:
@@ -481,17 +503,10 @@ static int cgre_receive_netlink_msg(int sk_nl)
 	struct sockaddr_nl from_nla;
 	socklen_t from_nla_len;
 	struct nlmsghdr *nlh;
-	struct sockaddr_nl kern_nla;
 	struct cn_msg *cn_hdr;
 
-	kern_nla.nl_family = AF_NETLINK;
-	kern_nla.nl_groups = CN_IDX_PROC;
-	kern_nla.nl_pid = 1;
-	kern_nla.nl_pad = 0;
-
 	memset(buff, 0, sizeof(buff));
 	from_nla_len = sizeof(from_nla);
-	memcpy(&from_nla, &kern_nla, sizeof(from_nla));
 	recv_len = recvfrom(sk_nl, buff, sizeof(buff), 0,
 		(struct sockaddr *)&from_nla, &from_nla_len);
 	if (recv_len == ENOBUFS) {
@@ -501,6 +516,14 @@ static int cgre_receive_netlink_msg(int sk_nl)
 	if (recv_len < 1)
 		return 0;
 
+	if (from_nla_len != sizeof(from_nla)) {
+		flog(LOG_ERR, "Bad address size reading netlink socket");
+		return 0;
+	}
+	if (from_nla.nl_groups != CN_IDX_PROC
+	    || from_nla.nl_pid != 0)
+		return 0;
+
 	nlh = (struct nlmsghdr *)buff;
 	while (NLMSG_OK(nlh, recv_len)) {
 		cn_hdr = NLMSG_DATA(nlh);
@@ -549,9 +572,12 @@ static void cgre_receive_unix_domain_msg(int sk_unix)
 		cgroup_dbg("read error: %s\n", strerror(errno));
 		goto close;
 	}
-	if (cgre_store_unchanged_process(pid, flags))
-		goto close;
-
+	if (flags == CGROUP_DAEMON_CANCEL_UNCHANGE_PROCESS) {
+		cgre_remove_unchanged_process(pid);
+	} else {
+		if (cgre_store_unchanged_process(pid, flags))
+			goto close;
+	}
 	if (write(fd_client, CGRULE_SUCCESS_STORE_PID,
 			sizeof(CGRULE_SUCCESS_STORE_PID)) < 0) {
 		cgroup_dbg("write error: %s\n", strerror(errno));
@@ -647,6 +673,23 @@ static int cgre_create_netlink_socket_process_msg(void)
 		cgroup_dbg("listening sk_unix error: %s\n", strerror(errno));
 		goto close_and_exit;
 	}
+
+	/* change the owner */
+	if (chown(CGRULE_CGRED_SOCKET_PATH, socket_user, socket_group) < 0) {
+		cgroup_dbg("Error changing socket owner: %s\n",
+				strerror(errno));
+		goto close_and_exit;
+	}
+	cgroup_dbg("Socket %s owner successfully set to %d:%d\n",
+			CGRULE_CGRED_SOCKET_PATH, (int) socket_user,
+			(int) socket_group);
+
+	if (chmod(CGRULE_CGRED_SOCKET_PATH, 0660) < 0) {
+		cgroup_dbg("Error changing socket owner: %s\n",
+				strerror(errno));
+		goto close_and_exit;
+	}
+
 	FD_ZERO(&readfds);
 	FD_SET(sk_nl, &readfds);
 	FD_SET(sk_unix, &readfds);
@@ -679,9 +722,9 @@ static int cgre_create_netlink_socket_process_msg(void)
 	}
 
 close_and_exit:
-	if (sk_nl)
+	if (sk_nl > 0)
 		close(sk_nl);
-	if (sk_unix)
+	if (sk_unix > 0)
 		close(sk_unix);
 	return rc;
 }
@@ -924,8 +967,11 @@ int main(int argc, char *argv[])
 	/* Return codes */
 	int ret = 0;
 
+	struct passwd *pw;
+	struct group *gr;
+
 	/* Command line arguments */
-	const char *short_options = "hvqf:s::ndQ";
+	const char *short_options = "hvqf:s::ndQu:g:";
 	struct option long_options[] = {
 		{"help", no_argument, NULL, 'h'},
 		{"verbose", no_argument, NULL, 'v'},
@@ -935,6 +981,8 @@ int main(int argc, char *argv[])
 		{"nodaemon", no_argument, NULL, 'n'},
 		{"debug", no_argument, NULL, 'd'},
 		{"nolog", no_argument, NULL, 'Q'},
+		{"socket-user", required_argument, NULL, 'u'},
+		{"socket-group", required_argument, NULL, 'g'},
 		{NULL, 0, NULL, 0}
 	};
 
@@ -1000,7 +1048,28 @@ int main(int argc, char *argv[])
 			verbosity = 4;
 			logp = "-";
 			break;
-
+		case 'u': /* --socket-user */
+			pw = getpwnam(optarg);
+			if (pw == NULL) {
+				usage(stderr, "Cannot find user %s", optarg);
+				ret = 3;
+				goto finished;
+			}
+			socket_user = pw->pw_uid;
+			cgroup_dbg("Using socket user %s id %d\n",
+					optarg, (int)socket_user);
+			break;
+		case 'g': /* --socket-group */
+			gr = getgrnam(optarg);
+			if (gr == NULL) {
+				usage(stderr, "Cannot find group %s", optarg);
+				ret = 3;
+				goto finished;
+			}
+			socket_group = gr->gr_gid;
+			cgroup_dbg("Using socket group %s id %d\n",
+					optarg, (int)socket_group);
+			break;
 		default:
 			usage(stderr, "");
 			ret = 2;
@@ -1010,23 +1079,24 @@ int main(int argc, char *argv[])
 
 	/* Initialize libcgroup. */
 	if ((ret = cgroup_init()) != 0) {
-		fprintf(stderr, "Error: libcgroup initialization failed, %d\n",
-				ret);
+		fprintf(stderr, "Error: libcgroup initialization failed, %s\n",
+				cgroup_strerror(ret));
 		goto finished;
 	}
 
 	/* Ask libcgroup to load the configuration rules. */
 	if ((ret = cgroup_init_rules_cache()) != 0) {
 		fprintf(stderr, "Error: libcgroup failed to initialize rules"
-				"cache, %d\n", ret);
+				"cache from %s. %s\n", CGRULES_CONF_FILE,
+				cgroup_strerror(ret));
 		goto finished;
 	}
 
 	/* Now, start the daemon. */
 	ret = cgre_start_daemon(logp, facility, daemon, verbosity);
 	if (ret < 0) {
-		fprintf(stderr, "Error: Failed to launch the daemon, %d\n",
-			ret);
+		fprintf(stderr, "Error: Failed to launch the daemon, %s\n",
+			cgroup_strerror(ret));
 		goto finished;
 	}
 
