@@ -31,6 +31,20 @@
 
 #define TEMP_BUF	81
 
+static void usage(int status, const char *program_name)
+{
+	if (status != 0) {
+		fprintf(stderr, "Wrong input parameters,"
+			" try %s -h' for more information.\n",
+			program_name);
+	} else {
+		printf("usage is %s "
+			"[-g <list of controllers>:<relative path to cgroup>] "
+			"[--sticky | --cancel-sticky] <list of pids>  \n",
+			program_name);
+	}
+}
+
 /*
  * Change process group as specified on command line.
  */
@@ -45,10 +59,11 @@ static int change_group_path(pid_t pid, struct cgroup_group_spec *cgroup_list[])
 
 		ret = cgroup_change_cgroup_path(cgroup_list[i]->path, pid,
                                                 (const char*const*) cgroup_list[i]->controllers);
-		if (ret)
+		if (ret) {
 			fprintf(stderr, "Error changing group of pid %d: %s\n",
 				pid, cgroup_strerror(ret));
 			return -1;
+		}
 	}
 
 	return 0;
@@ -93,6 +108,7 @@ out:
 
 static struct option longopts[] = {
 	{"sticky", no_argument, NULL, 's'},
+	{"cancel-sticky", no_argument, NULL, 'u'},
 	{0, 0, 0, 0}
 };
 
@@ -101,22 +117,24 @@ int main(int argc, char *argv[])
 	int ret = 0, i, exit_code = 0;
 	pid_t pid;
 	int cg_specified = 0;
-	int flag_child = 0;
+	int flag = 0;
 	struct cgroup_group_spec *cgroup_list[CG_HIER_MAX];
 	int c;
+	char *endptr;
 
 
 	if (argc < 2) {
-		fprintf(stderr, "usage is %s "
-			"[-g <list of controllers>:<relative path to cgroup>] "
-			"[--sticky] <list of pids>  \n",
-			argv[0]);
+		usage(1, argv[0]);
 		exit(2);
 	}
 
 	memset(cgroup_list, 0, sizeof(cgroup_list));
-	while ((c = getopt_long(argc, argv, "+g:s", longopts, NULL)) > 0) {
+	while ((c = getopt_long(argc, argv, "+g:sh", longopts, NULL)) > 0) {
 		switch (c) {
+		case 'h':
+			usage(0, argv[0]);
+			exit(0);
+			break;
 		case 'g':
 			ret = parse_cgroup_spec(cgroup_list, optarg,
 					CG_HIER_MAX);
@@ -128,10 +146,13 @@ int main(int argc, char *argv[])
 			cg_specified = 1;
 			break;
 		case 's':
-			flag_child |= CGROUP_DAEMON_UNCHANGE_CHILDREN;
+			flag |= CGROUP_DAEMON_UNCHANGE_CHILDREN;
+			break;
+		case 'u':
+			flag |= CGROUP_DAEMON_CANCEL_UNCHANGE_PROCESS;
 			break;
 		default:
-			fprintf(stderr, "Invalid command line option\n");
+			usage(1, argv[0]);
 			exit(2);
 			break;
 		}
@@ -146,10 +167,17 @@ int main(int argc, char *argv[])
 	}
 
 	for (i = optind; i < argc; i++) {
-		pid = (uid_t) atoi(argv[i]);
+		pid = (uid_t) strtol(argv[i], &endptr, 10);
+		if (endptr[0] != '\0') {
+			/* the input argument was not a number */
+			fprintf(stderr, "Error: %s is not valid pid.\n",
+				argv[i]);
+			exit_code = 2;
+			continue;
+		}
 
-		if (flag_child)
-			ret = cgroup_register_unchanged_process(pid, flag_child);
+		if (flag)
+			ret = cgroup_register_unchanged_process(pid, flag);
 		if (ret)
 			exit_code = 1;
 
