@@ -13,6 +13,10 @@
  *
  */
 
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+
 #include <errno.h>
 #include <grp.h>
 #include <libcgroup.h>
@@ -31,10 +35,25 @@
 #include "tools-common.h"
 
 static struct option longopts[] = {
-	{"sticky", no_argument, NULL, 's'}, 
+	{"sticky", no_argument, NULL, 's'},
+	{"help", no_argument, NULL, 'h'},
 	{0, 0, 0, 0}
 };
 
+static void usage(int status, const char *program_name)
+{
+	if (status != 0)
+		fprintf(stderr, "Wrong input parameters,"
+			" try %s --help' for more information.\n",
+			program_name);
+	else {
+		printf("Usage: %s [-h] [-g <controllers>:<path>] "
+			"[--sticky] command [arguments] ...\n",
+			program_name);
+	}
+}
+
+
 int main(int argc, char *argv[])
 {
 	int ret = 0, i;
@@ -46,17 +65,9 @@ int main(int argc, char *argv[])
 	int c;
 	struct cgroup_group_spec *cgroup_list[CG_HIER_MAX];
 
-	if (argc < 2) {
-		fprintf(stderr, "Usage is %s"
-			" [-g <list of controllers>:<relative path to cgroup>]"
-			" [--sticky] command [arguments]  \n",
-			argv[0]);
-		exit(2);
-	}
-
 	memset(cgroup_list, 0, sizeof(cgroup_list));
 
-	while ((c = getopt_long(argc, argv, "+g:s", longopts, NULL)) > 0) {
+	while ((c = getopt_long(argc, argv, "+g:sh", longopts, NULL)) > 0) {
 		switch (c) {
 		case 'g':
 			ret = parse_cgroup_spec(cgroup_list, optarg,
@@ -71,16 +82,18 @@ int main(int argc, char *argv[])
 		case 's':
 			flag_child |= CGROUP_DAEMON_UNCHANGE_CHILDREN;
 			break;
+		case 'h':
+			usage(0, argv[0]);
+			exit(0);
 		default:
-			fprintf(stderr, "Invalid command line option\n");
+			usage(1, argv[0]);
 			exit(1);
-			break;
 		}
 	}
 
 	/* Executable name */
 	if (!argv[optind]) {
-		fprintf(stderr, "No command specified\n");
+		usage(1, argv[0]);
 		exit(1);
 	}
 
@@ -92,6 +105,11 @@ int main(int argc, char *argv[])
 		return ret;
 	}
 
+	/* Just for debugging purposes. */
+	uid = geteuid();
+	gid = getegid();
+	cgroup_dbg("My euid and eguid is: %d,%d\n", (int) uid, (int) gid);
+
 	uid = getuid();
 	gid = getgid();
 	pid = getpid();
@@ -105,10 +123,14 @@ int main(int argc, char *argv[])
 	/*
 	 * 'cgexec' command file needs the root privilege for executing
 	 * a cgroup_register_unchanged_process() by using unix domain
-	 * socket, and an euid should be changed to the executing user
+	 * socket, and an euid/egid should be changed to the executing user
 	 * from a root user.
 	 */
-	if (seteuid(uid)) {
+	if (setresuid(uid, uid, uid)) {
+		fprintf(stderr, "%s", strerror(errno));
+		return -1;
+	}
+	if (setresgid(gid, gid, gid)) {
 		fprintf(stderr, "%s", strerror(errno));
 		return -1;
 	}
